from nose.tools import eq_

from .. import pubmed


def test_extract():
    
    text = """
    This is some text with a template cite. {{cite|...|...|pmid=1}}.
    This is some text with a template cite. {{cite|...|...|pmid = 2|...}}.
    This is some text with a template cite. {{cite|...|...|pmc = 3|...}}.
    This is some text with a template cite. {{cite|...|...|pmc = pmc4|...}}.
    This is some text with a link [http://www.ncbi.nlm.nih.gov/pubmed/5 ID]
    Another link [https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6 ID]
    """
    ids = list(pubmed.extract(text))
    expected = [
        ('pmid', "1"),
        ('pmid', "2"),
        ('pmc', "3"),
        ('pmc', "4"),
        ('pmid', "5"),
        ('pmc', "6")
    ]
    print(ids)
    print(expected)
    eq_(ids, expected)



import re
from collections import defaultdict

import mwparserfromhell as mwp
from more_itertools import peekable

from ..identifier import Identifier

DOI_RE = re.compile(r'\b(10\.\d+/[^\s\|\]\}\?\,]+)')

def extract_regex(text):
    for match in DOI_RE.finditer(text):
        id = re.sub(TAGS_RE, "", match.group(1)).rstrip(".")
        yield Identifier("doi", id)

DOI_START_RE = re.compile(r'10\.[0-9]{4,}/')

HTML_TAGS = ['ref', 'span', 'div', 'table', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
             'b', 'u', 'i', 's', 'ins', 'del', 'code', 'tt', 'blockquote',
             'pre']

TAGS_RE = re.compile(r'<(/\s*)?(' + '|'.join(HTML_TAGS) + ')(\s[^>\n\r]+)?>', re.I)


def extract_mwp(text):
    no_tags = mwp.parse(text).strip_code()
    for match in DOI_RE.finditer(no_tags):
        id = re.sub(TAGS_RE, "", match.group(1)).rstrip(".")
        yield Identifier("doi", id)

LEXICON = [
    (DOI_START_RE.pattern, 'doi_start'),
    (r'\(',                'open_paren'),
    (r'\)',                'close_paren'),
    (r'\[',                'open_bracket'),
    (r'\]',                'close_bracket'),
    (r'<!--',              'comment_start'),
    (r'-->',               'comment_end'),
    (TAGS_RE.pattern,      'tag'),
    (r'<',                 'open_angle'),
    (r'>',                 'close_angle'),
    (r'\{',                'open_curly'),
    (r'\}',                'close_curly'),
    (r'\|',                'pipe'),
    (r'[,\.;!]',           'punct'),
    (r'[\?#]',             'url_end'),
    (r'[\n\r]+',           'break'),
    (r'\s+',               'whitespace'),
    (r'\w+',               'word'),
    (r'.',                 'etc')
]

def extract_island(text):
    tokens = tokenize_finditer(text, LEXICON)
    tokens = peekable(tokens)
    
    while tokens.peek(None) is not None:
        
        if tokens.peek()[0] == 'doi_start':
            yield ('doi', read_doi(tokens))
        
        next(tokens)


def tokenize_finditer(text, lexicon=LEXICON):
    pattern = '|'.join("(?P<{0}>{1})".format(name, pattern)
                       for pattern, name in lexicon)
    
    group_regex = re.compile(pattern, re.I|re.U|re.M)
    
    for match in group_regex.finditer(text):
        yield match.lastgroup, match.group(0)


"""
def tokenize_scanner(text, lexicon=LEXICON):
    scanner = re.Scanner(lexicon)
    tokens, remainder = scanner.scan(text)
    return tokens
"""

#from mwcites.extractors.doi import tokenize_scan
#list(tokenize_scan("foo bar baz.{}"))

def read_doi(tokens):
    assert tokens.peek()[0] == 'doi_start'
    
    depth = defaultdict(lambda: 0)
    
    doi_buffer = [next(tokens)[1]]
    
    while tokens.peek(None) is not None:
        name, match = tokens.peek()
        
        if name in ('url_end', 'break', 'whitespace', 'tag', 'pipe',
                    'comment_start', 'comment_end'):
            break
        elif name == 'open_bracket':
            depth['bracket'] += 1
            doi_buffer.append(next(tokens)[1])
        elif name == 'open_curly':
            depth['curly'] += 1
            doi_buffer.append(next(tokens)[1])
        elif name == 'close_bracket':
            if depth['bracket'] > 0:
                depth['bracket'] -= 1
                doi_buffer.append(next(tokens)[1])
            else:
                break
        elif name == 'close_curly':
            if depth['curly'] > 0:
                depth['curly'] -= 1
                doi_buffer.append(next(tokens)[1])
            else:
                break
        else:
            doi_buffer.append(next(tokens)[1])
        
    
    # Do not return a doi with punctuation at the end
    return re.sub(r'[\.,!]+$', '', ''.join(doi_buffer))



def tokenize_search(text, start, lexicon=LEXICON):
    pattern = '|'.join("(?P<{0}>{1})".format(name, pattern)
                       for pattern, name in lexicon)
    
    group_regex = re.compile(pattern, re.I|re.U)
    
    match = group_regex.search(text, start)
    while match is not None:
        yield match.lastgroup, match.group(0)
        match = group_regex.search(text, match.span()[1])

def extract_search(text, lexicon=LEXICON):
    
    last_end = 0
    for match in DOI_START_RE.finditer(text):
        if match.span()[0] > last_end:
            tokens = tokenize_search(text, match.span()[0], lexicon=lexicon)
            tokens = peekable(tokens)
            doi = read_doi(tokens)
            last_end = match.span()[0] + len(doi)
            yield Identifier('doi', doi)
        else:
            last_end = max(match.span()[1], last_end)

extract = extract_search # Setting the default to the best method



"""
Gathers (looks up) metadata for citations based on type.


Usage:
    lookup -h | --help
    lookup [--metadata-cache=<path>] [--max-cache-size=<val>]

Options:
    -h --help                Print this documentation
    --metadata-cache=<path>  The location of a file containing cached metadata
                             [default: ./cache.tsv]
    --max-cache-size=<val>   The maximum number of entries to keep in the cache
                             [default: 2000000]
"""
'''
import sys
from functools import lru_cache

import docopt

from ..sources import doi, pubmed

ALL_SOURCES = {
    'doi': doi.lookup,
    'pmid': pubmed.lookup,
    'pcmid': pubmed.lookup
}


def main(argv=sys.argv):
    args = docopt.docopt(__doc__, argv=argv)
    

def run(cites, sources, max_cache_size=2000000, metadata_cache):
    
    for cite, metadata in lookup(cites, sources, max_cache_size, metadata_cache):
        
        print("\t".join(encode(v) for v in cite + [json.dumps(metadata)]))
        

def lookup(cites, sources, max_cache_size, metadata_cache):
    
    def get_metadata(identifier):
        
        if identifier not in metadata_cache:
            metadata = sources[identifier.type].lookup(identifier)
            metadata_cache[identifier] = metadata
        
        return metadata_cache[identifier]
    
    for cite in cites:
        type, id = cite[4:6]
        identifier = Identifier(type, id)
        
        metadata = get_metadata(identifier)
        return cite, metadata


class MetadataCache(OrderedDict):
    
    def __init__(self, *args, **kwargs):
        self.maxsize = kwargs.pop("maxsize", None)
        OrderedDict.__init__(self, *args, **kwargs)
        self._check_size_limit()
    
    def __setitem__(self, key, value):
        OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()
    
    def update(self, other):
        raise NotImplementedError()
    
    def _check_size_limit(self):
        if self.maxsize is not None:
            while len(self) > self.maxsize:
                self.popitem(last=False)
            
    
    def load(self, f):
        
        for line in f:
            type, id, metadata = line.rstrip().split("\t")
            type, id, metadata = str(type), str(id), json.loads(metadata)
            
            identifier = Identifier(type, id)
            self[identifier] = metadata
            
    
    
    def dump(self, f):
        
        for identifier, metadata in self.items():
            
            line = "\t".join([identifier.type,
                              identifier.id,
                              json.dumps(metadata)])
            f.write(line + "\n")
'''



"""
This script provides access to a set of utilities for processing academic
citations in Wikipedia.

Right now, there's only one utility, but there will be more to come.

* extract -- Extracts citations from an XML database dump

Usage:
    mwcitations (-h | --help)
    mwcitations <utility> [-h | --help]
    
Options:
    -h | --help  Shows this documentation
    <utility>    The name of the utility to run
"""
import sys
import traceback
from importlib import import_module

import docopt


USAGE = """Usage:
    mwcitations (-h | --help)
    mwcitations <utility> [-h | --help]\n"""


def main():
    
    if len(sys.argv) < 2:
        sys.stderr.write(USAGE)
        sys.exit(1)
    elif sys.argv[1] in ("-h", "--help"):
        sys.stderr.write(__doc__ + "\n")
        sys.exit(1)
    elif sys.argv[1][:1] == "-":
        sys.stderr.write(USAGE)
        sys.exit(1)
    
    module_name = sys.argv[1]
    try:
        module = import_module(".utilities." + module_name, package="mwcites")
    except ImportError:
        sys.stderr.write(traceback.format_exc())
        sys.stderr.write("Could not find utility {0}.\n".format(module_name))
        sys.exit(1)
    
    module.main(sys.argv[2:])



from collections import namedtuple

from mw import Timestamp
from nose.tools import eq_

from ..extract_cites import process_page


def process_page():
    FakeRevision = namedtuple("Revision", ['id', 'timestamp', 'text'])
    
    FakeExtractor = namedtuple("Extractor", ['extract'])
    
    fake_page = [
        FakeRevision(1, Timestamp(1), "id1 id2"),
        FakeRevision(2, Timestamp(2), "id1 id3"),
        FakeRevision(3, Timestamp(3), "id1 id2 id3"),
        FakeRevision(4, Timestamp(4), "id1 id2 id4"),
        FakeRevision(5, Timestamp(5), "id1 id2 id4"),
    ]
    fake_page.id = 1
    fake_page.title = "Title"
    
    extractor = FakeExtractor(lambda t: ('fake', id) for id in t.split(" "))
    
    cites = list(process_page(fake_page, [extractor]))
    
    eq_(cites,
        [(1, "Title", 1, Timestamp(1), "fake", "id1"),
         (1, "Title", 1, Timestamp(1), "fake", "id2"),
         (1, "Title", 4, Timestamp(4), "fake", "id4")])



"""
Extracts academic citations from articles from the history of Wikipedia articles
by processing a pages-meta-history XML dump and matching regular expressions
to revision content.

Currently supported identifies include:

 * PubMed
 * DOI
 
Outputs a TSV file with the following fields:

 * page_id: The identifier of the Wikipedia article (int), e.g. 1325125
 * page_title: The title of the Wikipedia article (utf-8), e.g. Club cell
 * rev_id: The Wikipedia revision where the citation was first added (int),
           e.g. 282470030
 * timestamp: The timestamp of the revision where the citation was first added.
              (ISO 8601 datetime), e.g. 2009-04-08T01:52:20Z
 * type: The type of identifier, e.g. pmid
 * id: The id of the cited scholarly article (utf-8),
       e.g 10.1183/09031936.00213411

Usage:
    extract -h | --help
    extract <dump_file>... [--extractor=<classpath>...]

Options:
    -h --help                Shows this documentation
    <dump_file>              The path to a set of dump files to process.  If no
                             files are specified, <stdin> will be read.
    --extractor=<classpath>  The class path to set of extractors to apply
                             [default: <all>]
"""
import sys
from itertools import chain

import docopt
from mw import xml_dump

from ..extractors import doi, pubmed

ALL_EXTRACTORS = [doi, pubmed]

HEADERS = ("page_id", "page_title", "rev_id", "timestamp", "type", "id")

def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    dump_files = args['<dump_file>']
    
    if args['--extractor'] == ['<all>']:
        extractors = ALL_EXTRACTORS
    else:
        extractors = [import_from_path(path) for path in args['--extractor']]
    
    run(dump_files, extractors)

def run(dump_files, extractors):
    
    print("\t".join(HEADERS))
    
    cites = extract(dump_files, extractors=extractors)
    for page_id, title, rev_id, timestamp, type, id in cites:
        
        print("\t".join(tsv_encode(v) for v in (page_id,
                                                title,
                                                rev_id,
                                                timestamp.long_format(),
                                                type,
                                                id)))

def extract(dump_files, extractors=ALL_EXTRACTORS):
    """
    Extracts cites from a set of `dump_files`.
    
    :Parameters:
        dump_files : str | `file`
            A set of files MediaWiki XML dump files
            (expects: pages-meta-history)
        extractors : `list`(`extractor`)
            A list of extractors to apply to the text
    
    :Returns:
        `iterable` -- a generator of extracted cites
    
    """
    # Dump processor function
    def process_dump(dump, path):
        for page in dump:
            if page.namespace != 0: continue
            else:
                for cite in extract_cite_history(page, extractors):
                    yield cite
        
    # Map call
    return xml_dump.map(dump_files, process_dump)

def extract_cite_history(page, extractors):
    """
    Extracts cites from the history of a `page` (`mw.xml_dump.Page`).
    
    :Parameters:
        page : `iterable`(`mw.xml_dump.Revision`)
            The page to extract cites from
        extractors : `list`(`extractor`)
            A list of extractors to apply to the text
    
    :Returns:
        `iterable` -- a generator of extracted cites
    
    """
    appearances = {} # For tracking the first appearance of an ID
    ids = set() # For holding onto the ids in the last revision.
    for revision in page:
        ids = set(extract_ids(revision.text, extractors))
        
        # For each ID, check to see if we have seen it before
        for id in ids:
            if id not in appearances:
               appearances[id] = (revision.id, revision.timestamp)
        
    for id in ids: #For the ids in the last version of the page
        rev_id, timestamp = appearances[id]
        yield (page.id, page.title, rev_id, timestamp, id.type, id.id)

def extract_ids(text, extractors):
    """
    Uses `extractors` to extract citation identifiers from a text.
    
    :Parameters:
        text : str
            The text to process
        extractors : `list`(`extractor`)
            A list of extractors to apply to the text
    
    :Returns:
        `iterable` -- a generator of extracted identifiers
    """
    for extractor in extractors:
        for id in extractor.extract(text):
            yield id

def import_from_path(path):
    """
    Imports a specific attribute from a module based on a class path.
    
    :Parameters:
        path : str
            A dot delimited string representing the import path of the desired
            object.
    
    :Returns:
        object -- An imported object
    """
    parts = path.split(".")
    module_path = ".".join(parts[:-1])
    attribute_name = parts[-1]

    module = import_module(module_path)

    attribute = getattr(module, attribute_name)

    return attribute


def tsv_encode(val, none_string="NULL"):
    """
    Encodes a value for inclusion in a TSV.  Basically, it converts the value
    to a string and escapes TABs and linebreaks.
    
    :Parameters:
        val : `mixed`
            The value to encode
        none_string : str
            The string to use when `None` is encountered
    
    :Returns:
        str -- a string representing the encoded value
    """
    if val == "None":
        return null_string
    else:
        if isinstance(val, bytes):
            val = str(val, 'utf-8')
        
        return str(val).replace("\t", "\\t").replace("\n", "\\n")



import re

from ..identifier import Identifier

TEMPLATE_RE = re.compile(r"\b(pmid|pmc)\s*=\s*(pmc)?([0-9]+)\b", re.I)

PMURL_RE = re.compile(r"//www\.ncbi\.nlm\.nih\.gov" +
                      r"/pubmed/([0-9]+)\b", re.I)
PMCURL_RE = re.compile(r"//www\.ncbi\.nlm\.nih\.gov" +
                       r"/pmc/articles/PMC([0-9]+)\b", re.I)

def extract(text):
    text = str(text or "")
    
    for match in TEMPLATE_RE.finditer(text):
        yield Identifier(match.group(1).lower(), match.group(3))
            
    for match in PMURL_RE.finditer(text):
        yield Identifier("pmid", match.group(1))
    
    for match in PMCURL_RE.finditer(text):
        yield Identifier("pmc", match.group(1))



import pprint

from nose.tools import eq_

from .. import doi
from ...identifier import Identifier

INPUT_TEXT = """
This is a doi randomly placed in the text 10.0000/m1
Here's a typo that might be construed as a doi 10.60 people were there.
{{cite|...|doi=10.0000/m2|pmid=10559875}}
<ref>Halfaker, A., Geiger, R. S., Morgan, J. T., & Riedl, J. (2012).
The rise and decline of an open collaboration system: How Wikipedia’s
reaction to popularity is causing its decline.
American Behavioral Scientist,
0002764212469365 doi: 10.1177/0002764212469365</ref>.  Hats pants and banana
[http://dx.doi.org/10.1170/foo<bar>(herp)derp]
[http://dx.doi.org/10.1170/foo<bar>(herp)derp[waffles]]
{{cite|...|doi=10.1098/rspb.2008.1131|issue=1656}}
http://www.google.com/sky/#latitude=3.362&longitude=160.1238441&zoom=
10.2387/234310.2347/39423
<!--
    10.2387/234310.2347/39423-->
"""
EXPECTED = [
    Identifier('doi', "10.0000/m1"),
    Identifier('doi', "10.0000/m2"),
    Identifier('doi', "10.1177/0002764212469365"),
    Identifier('doi', "10.1170/foo<bar>(herp)derp"),
    Identifier('doi', "10.1170/foo<bar>(herp)derp[waffles]"),
    Identifier('doi', "10.1098/rspb.2008.1131"),
    Identifier('doi', "10.2387/234310.2347/39423"),
    Identifier('doi', "10.2387/234310.2347/39423")
]

"""
def test_extract_regex():
    ids = list(doi.extract_regex(INPUT_TEXT))
    pprint.pprint(ids)
    pprint.pprint(EXPECTED)
    eq_(ids, EXPECTED)

def test_extract_mwp():
    ids = list(doi.extract_mwp(INPUT_TEXT))
    pprint.pprint(ids)
    pprint.pprint(EXPECTED)
    eq_(ids, EXPECTED)
"""

def test_extract():
    ids = list(doi.extract(INPUT_TEXT))
    pprint.pprint(ids)
    pprint.pprint(EXPECTED)
    eq_(ids, EXPECTED)

def test_extract_island():
    ids = list(doi.extract_island(INPUT_TEXT))
    pprint.pprint(ids)
    pprint.pprint(EXPECTED)
    eq_(ids, EXPECTED)
    
def test_extract_search():
    ids = list(doi.extract_search(INPUT_TEXT))
    pprint.pprint(ids)
    pprint.pprint(EXPECTED)
    #pprint.pprint(list(doi.tokenize_finditer(INPUT_TEXT)))
    eq_(ids, EXPECTED)



import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def requirements(fname):
    return [line.strip()
            for line in open(os.path.join(os.path.dirname(__file__), fname))]

setup(
    name = "mwcites",
    version = read('VERSION').strip(),
    author = "Aaron Halfaker",
    author_email = "ahalfaker@wikimedia.org",
    description = "A collection of scripts and utilities for extracting " +
                  "citations to academic literature from Wikipedia's XML " +
                  "database dumps.",
    license = "MIT",
    url = "https://github.com/halfak/Extract-scholarly-article-citations-from-Wikipedia",
    packages=find_packages(),
    entry_points = {
        'console_scripts': [
            'mwcitations=mwcites.mwcitations:main'
        ],
    },
    long_description = read('README.rst'),
    install_requires = ['docopt'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering"
    ]
)
